"""Implements the uvc command processing."""

import sys
import os
import subprocess
import tempfile
import logging
from cStringIO import StringIO

from uvc import commands, hg, svn
from uvc.util import run_in_directory
from uvc.path import path
from uvc.exc import *

log = logging.getLogger("uvc.main")

dialects = dict(hg=hg.HgDialect(), svn=svn.SVNDialect())

class IOWrapper(object):
    def __init__(self, file):
        self.file = file
        self.return_code = None
        
    def __getattr__(self, attr):
        return getattr(self.file, attr)

class Context(object):
    # all schemes are allowed
    remote_scheme_whitelist = None
    
    # a user can be set, because some DVCS operations
    # want a user to flag the commits, etc.
    user = None
    
    def __init__(self, working_dir, auth=None, timeout=-1,
                 output=None):
        """working_dir is the working directory in which commands should
        run. auth is a dictionary of authentication information:
        
        * type: 'ssh' or 'password'
        * key: for SSH, this should be the path to the user's private key
        * username: for SSH or password auth, this is the username to use
        * password: for password auth, this is the password to use
        
        timeout is the time (in seconds) that the command is allowed
        to run.
        
        output is a file-like object (specifically with a write method)
        to which the output is sent and return_code is set when
        the command is done. If it's not specified, it's set
        to sys.stdout.
        """
        self.working_dir = path(working_dir)
        self.auth = auth
        self.timeout = timeout
        self.output = IOWrapper(sys.stdout) if output is None else output
        
    def _normalize_path(self, unnorm_path):
        norm_path = self.working_dir / path(unnorm_path)
        norm_path = norm_path.normpath()
        norm_path = norm_path.realpath()
        norm_path = norm_path.abspath()
        return norm_path
    
    def validate_new_directory(self, newdir):
        """Confirms that newdir does not exist, raising a 
        FileError otherwise."""
        newdir = self._normalize_path(newdir)
        if newdir.exists():
            if newdir.isdir():
                raise FileError("Directory %s already exists" % newdir)
            else:
                raise FileError("Cannot create %s, path exists" % newdir)
        return newdir
    
    def validate_existing_directory(self, existingdir):
        """Confirms that existingdir does exist, raising a FileError
        otherwise."""
        existingdir = self._normalize_path(existingdir)
        if not existingdir.isdir():
            raise FileError("Directory %s does not exist" % existingdir)
        return existingdir
        
    def validate_new_file(self, newfile):
        """Confirms that newfile does not exist, raising a
        FileError otherwise."""
        newfile = self._normalize_path(newfile)
        if newfile.isfile():
            raise FileError("File %s already exists" % newfile)
        if newfile.isdir():
            raise FileError("%s exists as a directory" % newfile)
        return newfile
        
    def validate_existing_file(self, existingfile):
        """Confirms that existingfile exists, raising a FileError
        otherwise."""
        existingfile = self._normalize_path(existingfile)
        if not existingfile.isfile():
            raise FileError("File %s does not exist" % existingfile)
        return existingfile
        
    def validate_exists(self, existing):
        """Confirms that existing is a directory or
        file, raising a FileError otherwise."""
        existing = self._normalize_path(existing)
        if not existing.exists():
            raise FileError("File or directory %s does not exist" % (existing))
        return existing
        
    def validate_new_file_or_existing_dir(self, totest):
        """Confirms that totest either refers to a file that does
        not exist or a directory that does. This is useful for
        copy operations"""
        totest = self._normalize_path(totest)
        if not totest.exists():
            return totest
        if not totest.isdir():
            raise FileError("%s already exists" % totest)
        return totest
        
    def validate_path(self, path):
        """Validates that the path is legitimate. It doesn't
        need to point to anything existing, it just has to
        be able to pass _normalize_path."""
        return self._normalize_path(path)
    
class SecureContext(Context):
    # this is not the ideal place for this
    remote_scheme_whitelist = set(["http", "https", "ssh", "svn", "git", "bzr",
                                "svn+ssh"])
    
    def _normalize_path(self, unnorm_path):
        """Ensures that the path provided is underneath
        this context's working directory."""
        norm_path = super(SecureContext, self)._normalize_path(unnorm_path)
        rel_path = self.working_dir.relpathto(norm_path)
        if rel_path.isabs() or rel_path.startswith(".."):
            raise SecurityError("Path is outside of working directory: %s" 
                            % unnorm_path)
        return norm_path
    
def get_dialect(dialect_name):
    """Looks up the dialect in the dialect registry by name."""
    return dialects.get(dialect_name)

def infer_dialect(directory):
    d = set(dialects.values())
    cwd = os.getcwd()
    directory = os.path.abspath(directory)
    try:
        os.chdir(directory)
        remove_dialects = set()
        prev = None
        # stop when we hit the root directory (os.dirname
        # will stop giving us different values)
        while prev != directory:
            for dialect in d:
                is_match = dialect.cwd_is_this_dialect()
                if is_match == 0:
                    continue
                elif is_match == 1:
                    break
                else:
                    # in the case of Subversion, for example,
                    # if the directory doesn't have .svn in it, we
                    # know there's no match
                    remove_dialects.add(dialect)
            
            if is_match == 1:
                return dialect
            
            d.difference_update(remove_dialects)
            remove_dialects.clear()
            prev = directory
            directory = os.path.dirname(directory)
    finally:
        os.chdir(cwd)
    return None

def get_command_class(context, args, dialect=None):
    """This is similar to convert, but removes a step from the
    process. Use this if you need to inspect the command
    class before fulling setting up the context. Then,
    use from_args on the returned class to get an instance."""
    if dialect is None and args[0] in dialects:
        dialect_name = args.pop(0)
        dialect = get_dialect(dialect_name)
        command_name = args.pop(0)
        return dialect.get_dialect_command_class(command_name)
    else:
        generic_command_class = commands.get_command_class(context, args)
        guessed_dialect = generic_command_class.guess_dialect(context, args)
        if guessed_dialect:
            dialect = get_dialect(guessed_dialect)
        if dialect is None:
            raise commands.UVCError("Cannot run without knowing which VCS to use")
        return dialect.convert_class(generic_command_class)

def convert(context, args, dialect=None):
    """Converts the arguments provided into a command in the given dialect,
    or figures out the dialect for the project."""
    cmdclass = get_command_class(context, args, dialect)
    return cmdclass.from_args(context, args)

def run_command(command, context):
    command_line = command.get_command_line()
    
    # in some cases, such as Subversion's version of the
    # generic "commit", which is supposed to commit locally,
    # the command itself runs no vcs commands.
    if not command_line:
        return command.generate_output(context)
        
    log.debug("Running: %s", (command_line,))
    log.debug("Working dir: %s", context.working_dir)

    run_in_directory(context.working_dir, command_line, 
        timeout=context.timeout, 
        output=context.output)
    
    if context.output.return_code == 0 and \
        hasattr(command, "command_successful"):
        command.command_successful()
    
def _get_command_with_blanks_filled_in(context, args, dialect):
    """This will launch the user's text editor for
    items that really need to be filled in (such as
    commit messages)."""
    while True:
        try:
            return convert(context, args, dialect=dialect)
        except commands.GetValueFromEditor, e:
            editor = os.environ.get("EDITOR", "vi")
            tfile_handle, tfile_name = tempfile.mkstemp()
            try:
                command_line = [editor, tfile_name]
                p = subprocess.Popen(command_line).wait()
                contents = open(tfile_name).read()
            finally:
                os.unlink(tfile_name)
            if not contents:
                print "No text entered... exiting."
                sys.exit(1)
            args = e.template_args
            for i in range(0, len(args)):
                if args[i] == commands.inserted_value:
                    args[i] = contents
            
def is_new_project_command(args):
    log.debug("Verifying new project for %s", args)
    if args[0] == "clone" or args[0] == "checkout" \
       or (len(args) > 2 and " ".join(args[0:2]) in
       ["hg clone", "bzr clone", "git clone", "svn checkout"]):
        return True
    return False

def run(args=None):
    if args is None:
        args = sys.argv
    args.pop(0)
    
    cwd = os.getcwd()
    context = Context(cwd)
    
    if args and args[0] in dialects:
        dialect = None
    elif not is_new_project_command(args):
        dialect = infer_dialect(cwd)
    else:
        dialect = None
    
    command = _get_command_with_blanks_filled_in(context, args, dialect)
    run_command(command, context)

def call_uvc(directory, sub_command, params):
    """ Utility to call UVC from a given directory. This roughly the same as:
    $ cd $directory
    $ uvc $sub_command $params | $return
    'params' should be an array of params to pass into the sub_command.
    Returns an IOWrapper containing a StringIO with the output or containing the
    object passed in on the output_file parameter """
    output_file = StringIO()
    output = IOWrapper(output_file)
    context = SecureContext(directory, output=output)
    dialect = infer_dialect(directory);
    command_class = get_command_class(context, [ sub_command ], dialect)
    command = command_class.from_args(context, params)
    run_command(command, context)
    # TODO: This is a bit of a hack to allow call_uvc_data to get at the command
    output.command = command
    return output

def call_uvc_data(directory, sub_command, params):
    """ Utility to call UVC from a given directory. This roughly the same as:
    $ cd $directory
    $ uvc $sub_command $params | get_structured_data($return)
    'params' should be an array of params to pass into the sub_command.
    Returns output from command.get_structured_data rather than a string. """
    output = call_uvc(directory, sub_command, params)
    return output.command.get_structured_data(output.file.getvalue())



import os

from uvc import main, hg, commands, svn

topdir = os.path.abspath(os.path.join(os.path.dirname(__file__), "..", "..", 
                        "testfiles"))

context = None

def setup_module(module):
    global context
    if not os.path.exists(topdir):
        os.mkdir(topdir)
    context = main.Context(topdir)
    
def test_command_conversion():
    result = main.convert(context, ["hg", "clone", 
        "http://hg.mozilla.org/labs/bespin", "bespin"])
    assert isinstance(result, hg.clone)
    
def test_clone_command_can_infer_dialect():
    result = main.convert(context, ["clone", "http://hg.mozilla.org/labs/bespin", 
                          "bespin"])
    assert isinstance(result, hg.clone)
    
def test_is_new_project_command():
    true_tests = [
        "clone http://foo",
        "checkout http://foo",
        "hg clone http://foo",
        "svn checkout http://foo",
        "git clone http://foo",
        "bzr clone http://foo"
    ]
    
    false_tests = [
        "crone http://foo",
        "hg checkout http://foo"
    ]
    def run_one(to_test, expected):
        result = main.is_new_project_command(to_test)
        assert result == expected, "For %s, expected %s but got %s" % \
            (to_test, expected, result)
    
    for test in true_tests:
        yield run_one, test.split(), True
    for test in false_tests:
        yield run_one, test.split(), False
    
def test_get_command_class():
    result = main.get_command_class(context, ["hg", "commit"])
    assert result == hg.commit
    result = main.get_command_class(context, 
                ["clone", "http://hg.mozilla.org/labs/bespin"])
    assert result == hg.clone
    result = main.get_command_class(context,
                ["checkout", "http://foo/bar"], dialect=main.get_dialect('svn'))
    assert result == svn.clone



"""Implements the Subversion VCS dialect."""
import os
import re
import glob
from urlparse import urlsplit

import argparse

from uvc.commands import UVCError, DialectCommand, BaseCommand
from uvc.exc import RepositoryAlreadyInitialized, FileError
from uvc import util

class SVNError(UVCError):
    """A Subversion-specific error."""
    pass

class _DisplayHelp(Exception):
    pass
    
class _HelpAction(argparse.Action):
    """A replacement for the default help action that raises
    the DisplayHelp exception. This is done to ensure that the
    help output is directed to the correct place and then
    execution is stopped."""
    def __init__(self,
                 option_strings,
                 dest=argparse.SUPPRESS,
                 default=argparse.SUPPRESS,
                 help=None):
        super(_HelpAction, self).__init__(
            option_strings=option_strings,
            dest=dest,
            default=default,
            nargs=0,
            help=help)

    def __call__(self, parser, namespace, values, option_string=None):
        raise _DisplayHelp()
    
class OptionParser(argparse.ArgumentParser):
    def __init__(self, *args, **kw):
        kw['add_help'] = False
        kw['formatter_class'] = argparse.RawDescriptionHelpFormatter
        
        super(OptionParser, self).__init__(*args, **kw)
        
    def exit(self, status=0, msg=""):
        raise SVNError(msg)

class SVNCommand(DialectCommand):
    dialect_name = "svn"

def _add_auth_info(auth, parts):
    parts.insert(0, "--non-interactive")
    if auth:
        if auth['type'] == "ssh":
            parts.insert(1, "--trust-server-cert")
            parts.insert(2, "")
            parts.insert(3, "")
            parts.insert(4, "")
            #parts.insert(1, "-e")
            # setting StrictHostKeyChecking to no is less than ideal...
            #parts.insert(2, "ssh -i %s -o StrictHostKeyChecking=no" % (auth['key']))
        elif auth['type'] == "password":
            parts.insert(1, "--username")
            parts.insert(2, auth['username'])
            parts.insert(3, "--password")
            parts.insert(4, auth['password'])
            parts.insert(5, '--no-auth-cache')
    

class AuthSVNCommand(SVNCommand):
    def add_auth_info(self, parts):
        _add_auth_info(self.generic.auth, parts)
    
    def command_parts(self):
        parts = super(AuthSVNCommand, self).command_parts()
        self.add_auth_info(parts)
        return parts

class clone(AuthSVNCommand):
    reads_remote = True
    writes_remote = False

    def command_parts(self):
        parts = ["checkout", self.source_without_auth, self.dest]
        
        self.add_auth_info(parts)
            
        return parts

checkout = clone

def _commit_log_file(working_dir):
    return working_dir / ".svn_commit_messages"

class commit(SVNCommand):
    reads_remote = False
    writes_remote = False
    
    def command_parts(self):
        return []
        
    def get_command_line(self):
        return None
        
    def generate_output(self, context):
        commit_log = _commit_log_file(self.generic.working_dir)
        if not commit_log.exists():
            content = self.generic.message + "\n\n"
        else:
            content = commit_log.text() + self.generic.message + "\n\n"
        commit_log.write_text(content)
        context.output.write("Commit message saved. Don't forget to push to save to the remote repository!")
        context.output.return_code = 0

class diff(SVNCommand):
    reads_remote = False
    writes_remote = False

class remove(SVNCommand):
    reads_remote = False
    writes_remote = False
    
    def command_parts(self):
        parts = super(remove, self).command_parts()
        # uvc's remove command implies "force"
        parts.insert(1, "--force")
        return parts

delete = remove

_status_line_mask = re.compile("^(.)..... (.*)$")

def _get_file_list(working_dir, status="?"):
    retcode, stdout = util.run_in_directory(working_dir,
        ["svn", "status"])
    status_lines = stdout.read().split("\n")
    file_list = []
    for line in status_lines:
        m = _status_line_mask.match(line)
        if not m:
            continue
        if m.group(1) == status:
            filename = m.group(2)
            if not filename.startswith("."):
                file_list.append(m.group(2))
    
    return file_list

class add(SVNCommand):
    reads_remote = False
    writes_remote = False

    def command_parts(self):
        parts = super(add, self).command_parts()
        if not self.targets:
            parts.extend(_get_file_list(self.generic.working_dir))
        return parts

class push(AuthSVNCommand):
    reads_remote = True
    writes_remote = True
    
    def get_command_line(self):
        commit_log = _commit_log_file(self.generic.working_dir)
        if not commit_log.exists():
            return None
        return super(push, self).get_command_line()
        
    def generate_output(self, context):
        # only called if the commit_log doesn't exist
        output = context.output
        output.write("Nothing to push. Run commit first.")
        output.return_code = 1
    
    def command_parts(self):
        parts = []
        self.add_auth_info(parts)
        parts.append("commit")
        parts.append("-m")
        commit_log = _commit_log_file(self.generic.working_dir)
        parts.append(commit_log.text())
        return parts
        
    def command_successful(self):
        commit_log = _commit_log_file(self.generic.working_dir)
        commit_log.unlink()

class update(AuthSVNCommand):
    reads_remote = True
    writes_remote = False

    def command_parts(self):
        parts = ["update"]
        self.add_auth_info(parts)
        return parts

class resolved(SVNCommand):
   reads_remote = False
   writes_remote = False

   def command_parts(self):
       parts = super(resolved, self).command_parts()
       parts[0] = "resolve"
       parts.insert(1, "--accept")
       parts.insert(2, "working")
       if not self.targets:
           parts.extend(_get_file_list(self.generic.working_dir, status="C"))
       return parts

class status(SVNCommand):
    reads_remote = False
    writes_remote = False

class log(SVNCommand):
    reads_remote = False
    writes_remote = False

class cat(SVNCommand):
    reads_remote = False
    writes_remote = False

class revert(SVNCommand):
    reads_remote = False
    writes_remote = False

    def command_parts(self):
        parts = super(revert, self).command_parts()
        if not self.targets:
            parts.append("-R")
            pwd = os.getcwd()
            try:
                os.chdir(self.generic.working_dir)
                parts.extend(glob.glob("*"))
            finally:
                os.chdir(pwd)
        return parts
        
class SVNNativeCommand(object):
    existing_files = None
    aliases = []
    
    @classmethod
    def from_args(cls, context, initial_args):
        try:
            values = cls.parser.parse_args(initial_args)
            values.help = False
        except _DisplayHelp:
            values = argparse.Namespace(help=True)
            
        return cls(context, initial_args, values)
        
    @classmethod
    def name(cls):
        return cls.__name__[4:]
        
    def validate_args(self, context, values):
        """Validate the command line arguments passed in.
        
        Default is a no-op"""
        pass
        
    def __init__(self, context, initial_args, values):
        self.context = context
        self.auth = context.auth
        self.initial_args = initial_args
        self.values = values
        
        if values.help:
            return
            
        if self.existing_files:
            for var in self.existing_files:
                for f in getattr(self.values, var):
                    self.context.validate_exists(f)
                    
        self.validate_args(context, values)
        
        
    def command_parts(self):
        parts = [self.__class__.__name__[4:]] + self.initial_args
        _add_auth_info(self.auth, parts)
        return parts
        
    def get_command_line(self):
        if self.values.help:
            return None
        
        return ["svn"] + self.command_parts()
    
    def generate_output(self, context):
        self.parser.print_help(context.output)
        context.output.return_code = 0
    
def add_std(parser, cl=True, depth=True, files=True,
    recursive=False, quiet=True, xml=False, revision=False,
    verbose=False, force=False, parents=False, message=False, 
    accept=False):
    parser.add_argument("--help", "-h", action=_HelpAction,
                    default=argparse.SUPPRESS,
                    help="show this message and exit")
    
    if revision:
         parser.add_argument("--revision", "-r", help="""ARG (some commands also take ARG1:ARG2 range)
        A revision argument can be one of:
           NUMBER       revision number
           '{' DATE '}' revision at start of the date
           'HEAD'       latest in repository
           'BASE'       base rev of item's working copy
           'COMMITTED'  last commit at or before BASE
           'PREV'       revision just before COMMITTED""")
    
    if quiet:
        parser.add_argument("--quiet", "-q", action="store_true",
            help="print nothing, or only summary information")
    
    if verbose:
        parser.add_argument("--verbose", "-v", action="store_true",
            help="print extra information")

    if message:
        parser.add_argument("--message", "-m", 
            help="specify log message")
            
    if cl:
        parser.add_argument("--changelist", "--cl",
            help="operate only on members of changelist ARG")
            
    if depth:
        parser.add_argument("--depth", choices=["empty", "files",
                            "immediates", "infinity"],
                            help="limit operation by depth ARG ('empty', 'files', 'immediates', or 'infinity')")
                            
    if files:
        parser.add_argument("files", nargs='*',
            help="list of files to operate on")
    
    if recursive:
           parser.add_argument("--recursive", "-R", action='store_true',
               help="descend recursively, same as --depth=infinity")
    
    if xml:
        parser.add_argument("--xml", action="store_true",
            help="output in XML")
        parser.add_argument("--incremental", action="store_true",
            help="give output suitable for concatenation")
    
    if force:
        parser.add_argument("--force", action="store_true",
            help="force operation to run")
            
    if parents:
        parser.add_argument("--parents", action="store_true",
            help="make intermediate directories")
            
    if accept:
        parser.add_argument("--accept", 
            choices=('base', 'working', 'mine-full', 'theirs-full'),
            help="specify automatic conflict resolution source")
    
class SVN_revert(SVNNativeCommand):
    existing_files = ["files"]
    
    reads_remote = False
    writes_remote = False
    
    parser = OptionParser(prog="svn revert",
        description="revert: Restore pristine working copy file (undo most local edits).")
    add_std(parser, recursive=True)
        
class SVN_status(SVNNativeCommand):
    existing_files = ["files"]
    
    aliases = ["st", "stat"]
    
    reads_remote = False
    writes_remote = False
    
    parser = OptionParser(prog="svn status",
        description="""Print the status of working copy files and directories.

  With no args, print only locally modified items (no network access).
  With -q, print only summary information about locally modified items.
  With -u, add working revision and server out-of-date information.
  With -v, print full revision information on every item.

  The first six columns in the output are each one character wide:
    First column: Says if item was added, deleted, or otherwise changed
      ' ' no modifications
      'A' Added
      'C' Conflicted
      'D' Deleted
      'I' Ignored
      'M' Modified
      'R' Replaced
      'X' item is unversioned, but is used by an externals definition
      '?' item is not under version control
      '!' item is missing (removed by non-svn command) or incomplete
      '~' versioned item obstructed by some item of a different kind
    Second column: Modifications of a file's or directory's properties
      ' ' no modifications
      'C' Conflicted
      'M' Modified
    Third column: Whether the working copy directory is locked
      ' ' not locked
      'L' locked
    Fourth column: Scheduled commit will contain addition-with-history
      ' ' no history scheduled with commit
      '+' history scheduled with commit
    Fifth column: Whether the item is switched relative to its parent
      ' ' normal
      'S' switched
    Sixth column: Repository lock token
      (without -u)
      ' ' no lock token
      'K' lock token present
      (with -u)
      ' ' not locked in repository, no lock token
      'K' locked in repository, lock toKen present
      'O' locked in repository, lock token in some Other working copy
      'T' locked in repository, lock token present but sTolen
      'B' not locked in repository, lock token present but Broken

  The out-of-date information appears in the eighth column (with -u):
      '*' a newer revision exists on the server
      ' ' the working copy is up to date

  Remaining fields are variable width and delimited by spaces:
    The working revision (with -u or -v)
    The last committed revision and last committed author (with -v)
    The working copy path is always the final field, so it can
      include spaces.

  Example output:
    svn status wc
     M     wc/bar.c
    A  +   wc/qax.c

    svn status -u wc
     M           965    wc/bar.c
           *     965    wc/foo.c
    A  +         965    wc/qax.c
    Status against revision:   981

    svn status --show-updates --verbose wc
     M           965       938 kfogel       wc/bar.c
           *     965       922 sussman      wc/foo.c
    A  +         965       687 joe          wc/qax.c
                 965       687 joe          wc/zig.c
    Status against revision:   981
""")
    add_std(parser, xml=True, verbose=True)
    parser.add_argument("--show-updates", "-u", action="store_true",
        help="display update information")
    parser.add_argument("--ignore-externals", action="store_true",
        help="ignore externals definitions")

class SVN_add(SVNNativeCommand):
    existing_files = ["files"]
    
    reads_remote = False
    writes_remote = False
    
    parser = OptionParser(prog="svn add",
        description="""add: Put files and directories under version control, scheduling
them for addition to repository.  They will be added in next commit.""")
    add_std(parser)
    parser.add_argument("--force", action="store_true",
        help="force operation to run")
    parser.add_argument("--no-ignore", action="store_true",
        help="disregard default and svn:ignore property ignores")
    parser.add_argument("--auto-props", action="store_true",
        help="enable automatic properties")
    parser.add_argument("--no-auto-props", action="store_true",
        help="disable automatic properties")
    parser.add_argument("--parents", action="store_true",
        help="add intermediate parents")
        
class SVN_commit(SVNNativeCommand):
    existing_files = ["files"]
    
    aliases = ["ci"]
    
    reads_remote = True
    writes_remote = True
    
    parser = OptionParser(prog="svn commit",
        description="commit (ci): Send changes from your working copy to the repository.")
    add_std(parser, message=True)
    parser.add_argument("--no-unlock", action="store_true",
        help="don't unlock the targets")
    parser.add_argument("--with-revprop",
        help="set revision property ARG in new revision using the name[=value] format")
    parser.add_argument("--keep-changelists", action="store_true",
        help="don't delete changelists after commit")
        
class SVN_cleanup(SVNNativeCommand):
    existing_files = ["files"]
    
    reads_remote = False
    writes_remote = False
    
    parser = OptionParser(prog="svn cleanup",
        description="cleanup: Recursively clean up the working copy, removing locks, resuming unfinished operations, etc.")
    add_std(parser, cl=False, quiet=False, depth=False)
    

class SVN_update(SVNNativeCommand):
    existing_files = ["files"]
    aliases = ["up"]
    
    reads_remote = True
    writes_remote = False
    
    parser = OptionParser(prog="svn update",
        description="""update (up): Bring changes from the repository into the working copy.
        
    If no revision given, bring working copy up-to-date with HEAD rev.
    Else synchronize working copy to revision given by -r.

    For each updated item a line will start with a character reporting the
    action taken.  These characters have the following meaning:

      A  Added
      D  Deleted
      U  Updated
      C  Conflict
      G  Merged
      E  Existed

    A character in the first column signifies an update to the actual file,
    while updates to the file's properties are shown in the second column.
    A 'B' in the third column signifies that the lock for the file has
    been broken or stolen.

    If --force is used, unversioned obstructing paths in the working
    copy do not automatically cause a failure if the update attempts to
    add the same path.  If the obstructing path is the same type (file
    or directory) as the corresponding path in the repository it becomes
    versioned but its contents are left 'as-is' in the working copy.
    This means that an obstructing directory's unversioned children may
    also obstruct and become versioned.  For files, any content differences
    between the obstruction and the repository are treated like a local
    modification to the working copy.  All properties from the repository
    are applied to the obstructing path.  Obstructing paths are reported
    in the first column with code 'E'.

    Use the --set-depth option to set a new working copy depth on the
    targets of this operation.  Currently, the depth of a working copy
    directory can only be increased (telescoped more deeply); you cannot
    make a directory more shallow.
""")
    add_std(parser, revision = True, force=True)
    parser.add_argument("--set-depth", choices=["empty", "files",
                          "immediates", "infinity"],
                        help="set new working copy depth")
    parser.add_argument("--ignore-externals", action="store_true",
        help="ignore externals definitions")
    parser.add_argument("--accept", 
        choices=('postpone', 'base', 'mine-full', 'theirs-full'),
        help="specify automatic conflict resolution action")

class SVN_info(SVNNativeCommand):
    existing_files = ["files"]
    
    reads_remote = False
    writes_remote = False
    
    parser = OptionParser(prog="svn info",
        description="""info: Display information about a local or remote item.

    Print information about the paths given.
""")
    add_std(parser, xml=True, revision=True)
    
class SVN_blame(SVNNativeCommand):
    existing_files = ["files"]
    
    aliases = ["praise", "annotate", "ann"]
    
    reads_remote = True
    writes_remote = False
    
    parser = OptionParser(prog="svn blame",
        description="""blame (praise, annotate, ann): Output the content of specified files with revision and author information in-line.""")
    add_std(parser, revision=True, xml=True, verbose=True, force=True)
    parser.add_argument("--use-merge-history", "-g", action="store_true",
        help="use/display additional information from merge history")
    
class SVN_log(SVNNativeCommand):
    existing_files = ["files"]
    
    reads_remote = True
    writes_remote = False
    
    parser = OptionParser(prog="svn log",
        description="""log: Show the log messages for a set of revision(s) and/or file(s).

  1. Print the log messages for a local PATH (default: '.').
     The default revision range is BASE:1.

  2. Print the log messages for the PATHs (default: '.') under URL.
     If specified, REV determines in which revision the URL is first
     looked up, and the default revision range is REV:1; otherwise,
     the URL is looked up in HEAD, and the default revision range is
     HEAD:1.

  With -v, also print all affected paths with each log message.
  With -q, don't print the log message body itself (note that this is
  compatible with -v).

  Each log message is printed just once, even if more than one of the
  affected paths for that revision were explicitly requested.  Logs
  follow copy history by default.  Use --stop-on-copy to disable this
  behavior, which can be useful for determining branchpoints.

  Examples:
    svn log
    svn log foo.c
""")
    add_std(parser, xml=True, revision=True, verbose=True)
    parser.add_argument("--use-merge-history", "-g", action="store_true",
        help="use/display additional information from merge history")
    parser.add_argument("--change", "-c",
        help="the change made by CHANGE")
    parser.add_argument("--stop-on-copy", action="store_true",
        help="do not cross copies while traversing history")
    parser.add_argument("--limit", "-l", type=int,
        help="maximum number of log entries")
    parser.add_argument("--with-all-revprops", action="store_true",
        help="retrieve all revision properties")
    parser.add_argument("--with-revprop",
        help="retrieve revision property WITH_REVPROP")
    
class SVN_resolve(SVNNativeCommand):
    existing_files = ["files"]
    
    reads_remote = False
    writes_remote = False
    
    parser = OptionParser(prog="svn resolve",
        description="Resolve conflicts on working copy files or directories. (--accept is required)")
    
    add_std(parser, accept=True)
        
def _check_url(url_string):
    if " " in url_string:
        raise FileError("Spaces not allowed in URLs: %s" % (url_string))
    url = urlsplit(url_string)
    scheme = url[0].lower()
    if scheme not in ["http", "https", "svn+ssh"]:
        raise FileError("Unsupported URL: %s" % (url_string))
    

def _validate_urls_and_files(context, values):
    if len(values.src) == 1 and values.src[0].find("://") > -1:
        _check_url(values.src[0])
        if values.dst.find("://") > -1:
            _check_url(values.dst)
        else:
            context.validate_new_file_or_existing_dir(values.dst)
        return
        
    for f in values.src:
        context.validate_existing_file(f)
    
    if len(values.src) == 1:
        context.validate_new_file_or_existing_dir(values.dst)
    else:
        context.validate_existing_directory(values.dst)
    

class SVN_copy(SVNNativeCommand):
    reads_remote = True
    writes_remote = True
    
    aliases = ["cp"]
    
    parser = OptionParser(prog="svn copy",
        description="""copy (cp): Duplicate something in working copy or repository, remembering history.

When copying multiple sources, they will be added as children of DST,
which must be a directory.

  SRC and DST can each be either a working copy (WC) path or URL:
    WC  -> WC:   copy and schedule for addition (with history)
    WC  -> URL:  immediately commit a copy of WC to URL
    URL -> WC:   check out URL into WC, schedule for addition
    URL -> URL:  complete server-side copy;  used to branch and tag
  All the SRCs must be of the same type.

WARNING: For compatibility with previous versions of Subversion,
copies performed using two working copy paths (WC -> WC) will not
contact the repository.  As such, they may not, by default, be able
to propagate merge tracking information from the source of the copy
to the destination.
""")
    add_std(parser, files=False, revision=True, depth=False,
            cl=False, parents=True, message=True)
    parser.add_argument("--with-revprop",
        help="set revision property WITH_REVPROP in new revision using the name[=value] format")
    
    parser.add_argument("src", nargs="+", help="source file(s) or URL")
    parser.add_argument("dst", help="destination file, url or directory")
    
    def validate_args(self, context, values):
        _validate_urls_and_files(context, values)
    
class SVN_move(SVNNativeCommand):
    reads_remote = True
    writes_remote = True
    
    aliases = ["mv", "rename", "ren"]
    
    parser = OptionParser(prog="svn move",
        description="""Move and/or rename something in working copy or repository.

When moving multiple sources, they will be added as children of DST,
which must be a directory.

  Note:  this subcommand is equivalent to a 'copy' and 'delete'.
  Note:  the --revision option has no use and is deprecated.

  SRC and DST can both be working copy (WC) paths or URLs:
    WC  -> WC:   move and schedule for addition (with history)
    URL -> URL:  complete server-side rename.
  All the SRCs must be of the same type.
""")

    add_std(parser, files=False, depth=False, force=True,
            cl=False, parents=True, message=True)
    parser.add_argument("--with-revprop",
        help="set revision property WITH_REVPROP in new revision using the name[=value] format")
    parser.add_argument("src", nargs="+", help="source file(s) or URL(s)")
    parser.add_argument("dst", help="destination file, url or directory")

    def validate_args(self, context, values):
        _validate_urls_and_files(context, values)
    
class SVN_mkdir(SVNNativeCommand):
    reads_remote = True
    writes_remote = True
    
    parser = OptionParser(prog="svn mkdir",
        description="""Create a new directory under version control.

  Create version controlled directories.

  1. Each directory specified by a working copy PATH is created locally
    and scheduled for addition upon the next commit.

  2. Each directory specified by a URL is created in the repository via
    an immediate commit.

  In both cases, all the intermediate directories must already exist,
  unless the --parents option is given.
""")

    add_std(parser, files=False, parents=True, depth=False,
        message=True)
    
    parser.add_argument("path", nargs="+", help="URL or file path to create")
    
    def validate_args(self, context, values):
        for v in values.path:
            if v.find("://") > -1:
                _check_url(v)
            else:
                context.validate_new_directory(v)
    
class SVN_merge(SVNNativeCommand):
    reads_remote = True
    writes_remote = True
    
    parser = OptionParser(prog="svn merge",
        description="""Apply the differences between two sources to a working copy path.

usage: 1. merge sourceURL1[@N] sourceURL2[@M] [WCPATH]
       2. merge sourceWCPATH1@N sourceWCPATH2@M [WCPATH]
       3. merge [-c M[,N...] | -r N:M ...] SOURCE[@REV] [WCPATH]

  1. In the first form, the source URLs are specified at revisions
     N and M.  These are the two sources to be compared.  The revisions
     default to HEAD if omitted.

  2. In the second form, the URLs corresponding to the source working
     copy paths define the sources to be compared.  The revisions must
     be specified.

  3. In the third form, SOURCE can be either a URL or a working copy
     path (in which case its corresponding URL is used).  SOURCE (in
     revision REV) is compared as it existed between revisions N and M
     for each revision range provided.  If REV is not specified, HEAD
     is assumed.  '-c M' is equivalent to '-r <M-1>:M', and '-c -M'
     does the reverse: '-r M:<M-1>'.  If no revision ranges are
     specified, the default range of 0:REV is used.  Multiple '-c'
     and/or '-r' instances may be specified, and mixing of forward
     and reverse ranges is allowed.

  WCPATH is the working copy path that will receive the changes.
  If WCPATH is omitted, a default value of '.' is assumed, unless
  the sources have identical basenames that match a file within '.':
  in which case, the differences will be applied to that file.

  NOTE:  Subversion will only record metadata to track the merge
  if the two sources are on the same line of history -- if the
  first source is an ancestor of the second, or vice-versa.  This is
  guaranteed to be the case when using the third form listed above.
  The --ignore-ancestry option overrides this, forcing Subversion to
  regard the sources as unrelated and not to track the merge.

""")
    add_std(parser, files=False, revision=True, force=True, accept=True)
    
    parser.add_argument("--change", "-c", 
        help="the change made by revision ARG (like -r ARG-1:ARG) If ARG is negative this is like -r ARG:ARG-1")
    parser.add_argument("--dry-run", action="store_true",
        help="try operation but make no changes")
    parser.add_argument("--record-only", action="store_true",
        help="mark revisions as merged (use with -r)")
    parser.add_argument("--ignore-ancestry", 
        help="ignore ancestry when calculating merges")
    parser.add_argument("--reintegrate", action="store_true",
        help="lump-merge all of source URL's unmerged changes")
    
    parser.add_argument("paths", nargs="+")
    
    def validate_args(self, context, values):
        if values.change:
            if len(values.paths) > 2:
                raise SVNError("When using -c, only SOURCE[@REV] [WCPATH] arguments are permitted.")
                
            if "://" in values.paths[0]:
                _check_url(values.paths[0])
            else:
                context.validate_exists(values.paths[0])
            
            if len(values.paths) == 2:
                if "://" in values.paths[1]:
                    raise SVNError("The final argument must be a working copy path.")
                context.validate_exists(values.paths[1])
        else:
            if len(values.paths) < 2 or len(values.paths) > 3:
                raise SVNError("Two sources and an optional working copy path are required")
            if "://" in values.paths[0]:
                if "://" not in values.paths[1]:
                    raise SVNError("Both sources must be URLs")
                _check_url(values.paths[0])
                _check_url(values.paths[1])
            else:
                if "://" in values.paths[1]:
                    raise SVNError("Both sources must be working copy paths")
                context.validate_exists(values.paths[0])
                context.validate_exists(values.paths[1])
            
            if len(values.paths) == 3:
                context.validate_exists(values.paths[2])
     
class SVN_switch(SVNNativeCommand):
    reads_remote = True
    writes_remote = True
    
    aliases = ["sw"]
    
    parser = OptionParser(prog="svn switch",
        description="""Update the working copy to a different URL.

usage: 1. switch URL[@PEGREV] [PATH]
       2. switch --relocate FROM TO [PATH...]

  1. Update the working copy to mirror a new URL within the repository.
     This behaviour is similar to 'svn update', and is the way to
     move a working copy to a branch or tag within the same repository.
     If specified, PEGREV determines in which revision the target is first
     looked up.

     If --force is used, unversioned obstructing paths in the working
     copy do not automatically cause a failure if the switch attempts to
     add the same path.  If the obstructing path is the same type (file
     or directory) as the corresponding path in the repository it becomes
     versioned but its contents are left 'as-is' in the working copy.
     This means that an obstructing directory's unversioned children may
     also obstruct and become versioned.  For files, any content differences
     between the obstruction and the repository are treated like a local
     modification to the working copy.  All properties from the repository
     are applied to the obstructing path.

     Use the --set-depth option to set a new working copy depth on the
     targets of this operation.  Currently, the depth of a working copy
     directory can only be increased (telescoped more deeply); you cannot
     make a directory more shallow.

  2. Rewrite working copy URL metadata to reflect a syntactic change only.
     This is used when repository's root URL changes (such as a scheme
     or hostname change) but your working copy still reflects the same
     directory within the same repository.
""")
    add_std(parser, files=False, cl=False, revision=True, force=True, accept=True)
    
    parser.add_argument("--relocate", metavar="FROM",
        help="relocate via URL-rewriting")
    parser.add_argument("--ignore-externals", action="store_true",
        help="ignore externals definitions")
    
    parser.add_argument("url")
    parser.add_argument("path", nargs="?")
    
    def validate_args(self, context, values):
        if values.relocate:
            _check_url(values.relocate)
        
        _check_url(values.url)
        
        if values.path:
            context.validate_existing_directory(values.path)

class SVN_delete(SVNNativeCommand):
    reads_remote = True
    writes_remote = True
    
    aliases = ["del", "remove", "rm"]
    
    parser = OptionParser(prog="svn delete",
        description="""Remove files and directories from version control.

usage: 1. delete PATH...
       2. delete URL...

  1. Each item specified by a PATH is scheduled for deletion upon
    the next commit.  Files, and directories that have not been
    committed, are immediately removed from the working copy
    unless the --keep-local option is given.
    PATHs that are, or contain, unversioned or modified items will
    not be removed unless the --force option is given.

  2. Each item specified by a URL is deleted from the repository
    via an immediate commit.
""")
    add_std(parser, files=False, force=True, message=True, cl=False, 
        depth=False)
    parser.add_argument("--keep-local", action="store_true",
        help="keep path in working copy")
    parser.add_argument("files", nargs="+", metavar="PATH_OR_URL")
    
    def validate_args(self, context, values):
        for v in values.files:
            if "://" in v:
                _check_url(v)
            else:
                context.validate_exists(v)
            
class SVN_cat(SVNNativeCommand):
    reads_remote = True
    writes_remote = False
    
    parser = OptionParser(prog="svn cat",
        description="""Output the content of specified files or URLs.

usage: cat TARGET[@REV]...

  If specified, REV determines in which revision the target is first
  looked up.
""")
    add_std(parser, revision=True, files=False, depth=False, quiet=False,
                cl=False)
    parser.add_argument("targets", nargs="+")
    
    def validate_args(self, context, values):
        for v in values.targets:
            if "://" in v:
                _check_url(v)
            else:
                context.validate_path(v)

class SVN_propset(SVNNativeCommand):
    existing_files = ["path"]
    
    reads_remote = False
    writes_remote = False
    
    aliases = ["pset", "ps"]
    
    parser = OptionParser(prog="svn propset",
        description="""Set the value of a property on files, dirs, or revisions.

  Note: svn recognizes the following special versioned properties
  but will store any arbitrary properties set:
    svn:ignore     - A newline separated list of file patterns to ignore.
    svn:keywords   - Keywords to be expanded.  Valid keywords are:
      URL, HeadURL             - The URL for the head version of the object.
      Author, LastChangedBy    - The last person to modify the file.
      Date, LastChangedDate    - The date/time the object was last modified.
      Rev, Revision,           - The last revision the object changed.
      LastChangedRevision
      Id                       - A compressed summary of the previous
                                   4 keywords.
    svn:executable - If present, make the file executable.  Use
      'svn propdel svn:executable PATH...' to clear.
    svn:eol-style  - One of 'native', 'LF', 'CR', 'CRLF'.
    svn:mime-type  - The mimetype of the file.  Used to determine
      whether to merge the file, and how to serve it from Apache.
      A mimetype beginning with 'text/' (or an absent mimetype) is
      treated as text.  Anything else is treated as binary.
    svn:externals  - A newline separated list of module specifiers,
      each of which consists of a relative directory path, optional
      revision flags and an URL.  The ordering of the three elements
      implements different behavior.  Subversion 1.4 and earlier only
      support the following formats and the URLs cannot have peg
      revisions:
        foo             http://example.com/repos/zig
        foo/bar -r 1234 http://example.com/repos/zag
      Subversion 1.5 and greater support the above formats and the
      following formats where the URLs may have peg revisions:
                http://example.com/repos/zig foo
        -r 1234 http://example.com/repos/zig foo/bar
      Relative URLs are supported in Subversion 1.5 and greater for
      all above formats and are indicated by starting the URL with one
      of the following strings
        ../  to the parent directory of the extracted external
        ^/   to the repository root
        //   to the scheme
        /    to the server root
      The ambiguous format 'relative_path relative_path' is taken as
      'relative_url relative_path' with peg revision support.
    svn:needs-lock - If present, indicates that the file should be locked
      before it is modified.  Makes the working copy file read-only
      when it is not locked.  Use 'svn propdel svn:needs-lock PATH...'
      to clear.

  The svn:keywords, svn:executable, svn:eol-style, svn:mime-type and
  svn:needs-lock properties cannot be set on a directory.  A non-recursive
  attempt will fail, and a recursive attempt will set the property
  only on the file children of the directory.
""")
    parser.add_argument("propname")
    parser.add_argument("propval")
    parser.add_argument("path", nargs="+")
    add_std(parser, force=True, files=False)
    
class SVN_propget(SVNNativeCommand):
    existing_files = ["files"]
    
    reads_remote = False
    writes_remote = False
    
    aliases = ["pg", "pget"]
    
    parser = OptionParser(prog="svn propget",
        description="""Print the value of a property on files, dirs, or revisions.

  By default, this subcommand will add an extra newline to the end
  of the property values so that the output looks pretty.  Also,
  whenever there are multiple paths involved, each property value
  is prefixed with the path with which it is associated.  Use
  the --strict option to disable these beautifications (useful,
  for example, when redirecting binary property values to a file).
""")
    parser.add_argument("propname")
    add_std(parser, revision=True, quiet=False)
    parser.add_argument("--strict", action="store_true",
        help="use strict semantics")
    parser.add_argument("--xml", action="store_true",
        help="output in XML")
    
class SVN_propdel(SVNNativeCommand):
    existing_files = ["files"]
    
    reads_remote = False
    writes_remote = False
    
    aliases = ["pdel", "pd"]
    
    parser = OptionParser(prog="svn propdel",
        description="""Remove a property from files, dirs, or revisions.
""")
    
    parser.add_argument("propname")
    add_std(parser)
    
class SVN_proplist(SVNNativeCommand):
    existing_files = ["files"]
    
    reads_remote = False
    writes_remote = False
    aliases = ["plist", "pl"]
    
    parser = OptionParser(prog="svn proplist",
        description="""List all properties on files, dirs, or revisions.""")
    
    add_std(parser, verbose=True)
    parser.add_argument("--xml", action="store_true",
        help="output in XML")
    
class SVN_help(SVNNativeCommand):
    reads_remote=False
    writes_remote=False
    
    parser = OptionParser(prog="svn help",
        description="""Most subcommands take file and/or directory arguments, recursing on the directories.  If no arguments are supplied to such a command, it recurses on the current directory (inclusive) by default.""")
    
_native_commands = None
    
def get_native_commands():
    """Return a set of the command classes for all of the native
    SVN commands that are supported."""
    
    global _native_commands
    
    if _native_commands is not None:
        return _native_commands
        
    commands = []
    for key, val in globals().items():
        if key.startswith("SVN_"):
            commands.append(val)
        
    _native_commands = commands
    return commands
    
for command in get_native_commands():
    g = globals()
    for alias in command.aliases:
        g["SVN_" + alias] = command
    
class SVNDialect(object):
    
    name = "svn"
    
    def convert(self, command_object):
        """Converts to an SVN-specific command."""
        local_command = self._get_command_class(command_object.__class__.__name__)
        return local_command(command_object)
        
    def convert_class(self, command_class):
        """Converts a generic command class to its equivalent."""
        local_command_class = self._get_command_class(command_class.__name__)
        return local_command_class
        
    def _get_command_class(self, command_name):
        """Looks up an svn-specific command class."""
        local_command = globals().get(command_name)
        if not local_command:
            raise SVNError("Command cannot be converted to svn: %s" 
                            % (command_name))
        return local_command
        
    def get_dialect_command_class(self, command_name):
        """Looks up an svn-specific command class"""
        return self._get_command_class("SVN_" + command_name)
    
    def cwd_is_this_dialect(self):
        """Returns 1 if the .svn directory is here, 2 otherwise. svn
        plants directories everywhere, so if it's not in the current
        directory, it's not an svn project."""
        if os.path.isdir(".svn"):
            return 1
        return 2
    


import os
from cStringIO import StringIO

import py

from uvc.path import path
from uvc import commands, svn, main, exc, util
from uvc.tests.util import test_context
from uvc.tests.mock import patch

dialect = svn.SVNDialect()

topdir = path(__file__).dirname().abspath() / ".." / ".." / "testfiles"
testfile = topdir / "foo.txt"
testfile2 = topdir / "other.txt"
testdir = topdir / "adirectory"

context = None

def setup_module(module):
    global context
    if not os.path.exists(topdir):
        os.mkdir(topdir)
    context = main.Context(topdir)
    testfile.write_bytes("hi")
    testfile2.write_bytes("hello")
    testdir.mkdir()

def teardown_module(module):
    testfile.unlink()
    testdir.rmdir()
    testfile2.unlink()
    
def test_clone_command_conversion():
    generic_clone = commands.clone(context, ["http://paver.googlecode.com/svn/trunk/", 
                                    "paver"])
    svn_clone = svn.clone(generic_clone)
    svn_clone_command = " ".join(svn_clone.get_command_line())
    assert svn_clone_command == "svn --non-interactive checkout http://paver.googlecode.com/svn/trunk/ paver"
    
def test_convert_function():
    generic_clone = commands.clone(context, ["http://paver.googlecode.com/svn/trunk/", 
                                    "paver"])
    svn_clone = dialect.convert(generic_clone)
    svn_clone_command = " ".join(svn_clone.get_command_line())
    assert svn_clone_command == "svn --non-interactive checkout http://paver.googlecode.com/svn/trunk/ paver"
    
def test_convert_unknown_command():
    class Foo(object):
        pass
    
    try:
        result = dialect.convert(Foo())
        assert False, "expected SVNError for unknown command"
    except svn.SVNError:
        pass

def test_commit_command():
    test_context.reset()
    commit_file = path(test_context.working_dir) / ".svn_commit_messages"
    if commit_file.exists():
        commit_file.unlink()
    
    generic_commit = commands.commit(test_context, 
                ["-m", "test message", "foo", "bar"])
    result = dialect.convert(generic_commit)
    assert result.get_command_line() == None
    result.generate_output(test_context)
    output = test_context.output.getvalue()
    assert output == "Commit message saved. Don't forget to push to save to the remote repository!"
    
    assert commit_file.exists(), "Expected commit file at " + commit_file
    assert commit_file.text() == "test message\n\n"
    
def test_push_command_before_commit():
    test_context.reset()
    commit_file = path(test_context.working_dir) / ".svn_commit_messages"
    if commit_file.exists():
        commit_file.unlink()
    
    generic_push = commands.push(test_context, [])
    result = dialect.convert(generic_push)
    assert result.get_command_line() == None
    
    result.generate_output(test_context)
    output = test_context.output.getvalue()
    assert output == "Nothing to push. Run commit first."
    
def test_push_after_commit():
    commit_file = path(test_context.working_dir) / ".svn_commit_messages"
    if commit_file.exists():
        commit_file.unlink()
    
    generic_commit = commands.commit(test_context, 
                ["-m", "test message1", "foo", "bar"])
    result = dialect.convert(generic_commit)
    result.generate_output(context)
    
    generic_commit = commands.commit(test_context, 
                ["-m", "test message2", "foo", "bar"])
    result = dialect.convert(generic_commit)
    result.generate_output(context)
    
    generic_push = commands.push(test_context, [])
    result = dialect.convert(generic_push)
    assert result.get_command_line() == ["svn", 
        "--non-interactive", "commit", "-m", 
        "test message1\n\ntest message2\n\n"]
    
def test_diff_command():
    generic_diff = commands.diff(context, [])
    result = dialect.convert(generic_diff)
    assert str(result) == "diff"
    
@patch("uvc.util.run_in_directory")
def test_add_all_files(rid):
    generic_add = commands.add(test_context, [])
    result = dialect.convert(generic_add)
    stdout = StringIO("""A      foo.txt
?      bar.txt
A      baz.txt
""")
    rid.return_value = [0, stdout]
    assert result.get_command_line() == ["svn", "add", "bar.txt"]
    assert rid.called
    
def test_revert_all_files():
    generic_revert = commands.revert(context, [])
    result = dialect.convert(generic_revert)
    assert result.get_command_line() == ["svn", "revert", 
        "-R", "adirectory", "foo.txt", "other.txt"]
    
@patch("uvc.util.run_in_directory")
def test_resolve_all_files(rid):
    stdout = StringIO("""A      foo.txt
C      bar.txt
A      baz.txt
""")
    rid.return_value = [0, stdout]
    generic_resolved = commands.resolved(context, [])
    result = dialect.convert(generic_resolved)
    assert result.get_command_line() == ["svn", "resolve", 
        "--accept", "working", "bar.txt"]
    assert rid.called
    
class T(object):
    def __init__(self, command, is_help=False, 
                error=False, file_error=False):
        self.command = command
        self.error = error
        self.file_error = file_error
        self.is_help = is_help
        
    @property
    def args(self):
        return ["svn"] + self.command.split(" ")
        
    def __repr__(self):
        return "%s (is_help=%s, error=%s, file_error=%s)" % (
            self.command, self.is_help, self.error, self.file_error)
        
def test_svn_native_commands():
    tests = [
        T("revert foo.txt"),
        T("revert -R foo.txt"),
        T("revert --foo", error=True),
        T("revert -h", is_help=True),
        T("revert --help", is_help=True),
        T("revert bar.txt", file_error=True),
        T("revert --depth foo foo.txt", error=True),
        T("revert --depth=infinity foo.txt"),
        T("copy foo.txt bar.txt"),
        T("copy", error=True),
        T("copy foo.txt", error=True),
        T("copy bar.txt foo.txt", file_error=True),
        T("copy foo.txt adirectory"),
        T("copy other.txt adirectory foo.txt", file_error=True),
        T("copy foo.txt other.txt adirectory"),
        T("copy http://blazingthings.com/awesome/ http://blazingthings.com/awesome/a2"),
        T("copy file://foo/bar/baz .", file_error=True),
        T("copy http://www.blazingthings.com/awesome/@658 ."),
        T("copy http://www.blazingthings.com/awesome/@658 newdir"),
        T("move foo.txt bar.txt"),
        T("move http://www.blazingthings.com/awesome/blah http://www.blazingthings.com/awesome/bar"),
        T("move foo http://www.blazingthings.com/awesome/bar", file_error=True),
        T("mkdir testdir"),
        T("merge http://foobar.baz/branch1 http://foobar.baz/branch2 ."),
        T("merge http://foobar.baz/branch1 http://foobar.baz/branch2"),
        T("merge -c 10 http://foobar.baz/branch1"),
        T("merge http://foobar.baz/branch1 .", error=True),
        T("merge -c 10 http://foobar.baz/branch1 http://foobar.baz/branch2 .",
            error=True),
        T("merge -c 10 http://foobar.baz/branch1 http://foobar.baz/branch2",
            error=True),
        T("merge branch1 http://foobar.baz/branch1", error=True),
        T("switch http://foobar.baz/branch1"),
        T("switch http://foobar.baz/branch1 ."),
        T("switch http://foobar.baz/branch1 http://foobar.baz/branch2",
            file_error=True),
        T("switch .", file_error=True),
        T("switch --relocate http://foobar.baz/branch1 http://foobar.baz/branch2 ."),
        T("delete --keep-local foo.txt"),
        T("cat http://foobar.baz/file1"),
        T("cat http://foobar.baz/file1@329"),
        T("cat foo.txt@329"),
        T("cat file:///etc/passwd", file_error=True),
        T("propset foo.txt", error=True),
        T("propset svn:externals foo.txt", error=True),
        T("propset svn:externals http://blah foo.txt"),
        T("propset svn:externals http://blah DOESNTEXIST", 
            file_error=True),
        T("propget svn:ignore"),
        T("propget svn:mime-type foo.txt"),
        T("propget svn:mime-type DOESNOTEXIST", file_error=True),
        T("propdel svn:mime-type foo.txt"),
        T("propdel svn:mime-type DOESNOTEXIST", file_error=True),
        T("proplist"),
        T("proplist foo.txt"),
        T("proplist DOESNOTEXIST", file_error=True)
    ]
    
    def test_one(t):
        try:
            command = main.convert(context, t.args, None)
            if t.error or t.file_error:
                assert False, "For %s, expected error but didn't get one" % (t.args,)
        except svn.SVNError, e:
            if t.error:
                return
            else:
                assert False, "Unexpected Exception: %s" % e
        except exc.FileError, e:
            if t.file_error:
                return
            else:
                assert False, "Unexpected file error: %s" % e
                
        command_line = command.get_command_line()
        if command_line is None:
            return
        
        final_args = t.args
        final_args.insert(1, "--non-interactive")
        assert command_line == final_args
    
    for t in tests:
        yield test_one, t
    
def test_get_native_commands():
    commands = svn.get_native_commands()
    names = [command.name() for command in commands]
    assert "commit" in names
    



"""The basic set of Uber Version Controller commands. These are the commands
that are not specific to any given version control system."""

from optparse import OptionParser
from urlparse import urlparse, urlunparse

from uvc.path import path
from uvc.exc import *

class Sentinel(object):
    pass

inserted_value = Sentinel()

class GetValueFromEditor(UVCError):
    def __init__(self, template_args, prompt):
        self.template_args = template_args
        self.prompt = prompt
    
class DialectCommand(object):
    """Base class for the dialect-specific command classes.
    If you subclass this, the default behavior is to pass
    through the generic command interpretation."""
    
    # dialects should override this
    dialect_name = ""
    
    @classmethod
    def from_args(cls, context, args):
        generic = globals()[cls.__name__](context, args)
        return cls(generic)

    def __init__(self, generic):
        self.generic = generic

    def __str__(self):
        return " ".join(self.command_parts())
    
    def command_parts(self):
        return self.generic.command_parts()
    
    def get_command_line(self):
        return [self.dialect_name] + self.command_parts()
        
    def set_auth(context):
        """Updates the authentication information based on the
        context. It is generally best for __init__ to call
        this, but there will be times when the authentication
        requirements are not known before this object is
        created. The default behavior is to do nothing."""
        pass
        
    def __getattr__(self, attr):
        return getattr(self.generic, attr)
    
class BaseCommand(object):
    def __init__(self, context, args):
        self.working_dir = context.working_dir
        self.auth = context.auth
        self.user = context.user
    
    def __str__(self):
        return " ".join(self.command_parts())
    
    @classmethod
    def guess_dialect(cls, context, args):
        return None

def _apply_auth(url, context):
    parsed = urlparse(url)
    
    if parsed.scheme == "" or parsed.scheme == "file":
        # we don't need to apply auth info to local files
        return url
        
    auth = context.auth
    if auth:
        if 'username' in auth or 'password' in auth:
            # build up username/password for URL from the context
            new_url = list(parsed)
            auth_info = [auth['username']]
            if 'password' in auth:
                auth_info.append(auth['password'])
            
            # strip off the old auth info
            netloc = parsed.netloc
            if "@" in netloc:
                netloc = netloc.split('@')[1]
                
            new_url[1] = "%s@%s" % (":".join(auth_info), netloc)
            url = urlunparse(new_url)
    return url
    

class clone(BaseCommand):
    # source of the repository to get
    source = None
    
    # where to put the resulting files
    dest = None
    
    reads_remote = True
    writes_remote = False
    
    @classmethod
    def guess_dialect(cls, context, args):
        guessed = None
        source = args[0]
        if "hg" in source:
            guessed = "hg"
        elif "svn" in source:
            guessed = "svn"
        elif "git" in source:
            guessed = "git"
        elif "bzr" in source:
            guessed = "bzr"
        return guessed
    
    def __init__(self, context, args):
        super(clone, self).__init__(context, args)
        if len(args) < 1:
            raise BadArgument("Clone requires a source argument")
        source = args[0]
        
        self.guessed_dialect = self.guess_dialect(context, args)
        
        parsed = urlparse(source)
        
        self.source_without_auth = source
        
        source = _apply_auth(source, context)
                    
        self.source = source
        
        if len(args) == 2:
            self.dest = args[1]
        else:
            if parsed.path.endswith("/"):
                index = -2
            else:
                index = -1
            last_path = parsed.path.split('/')[index]
            if not last_path:
                raise BadArgument("Clone requires source and dest arguments.")
            self.dest = last_path

        context.validate_new_directory(self.dest)
    
    def command_parts(self):
        return ["clone", self.source, self.dest]
    
checkout = clone
    
class commit(BaseCommand):
    """Commit command."""
    # commit message
    message = None
    
    # specific directories/files to commit
    targets = None
    
    # this is the DVCS view of commit
    reads_remote = False
    writes_remote = False
    
    parser = OptionParser()
    parser.add_option('-m', "--message", dest="message", 
        help="commit message")
    
    def __init__(self, context, args):
        super(commit, self).__init__(context, args)
        options, args = self.parser.parse_args(args)
        if not options.message:
            args = ["commit", "-m", inserted_value] + args
            raise GetValueFromEditor(args, "Please enter a commit message")
        self.message = options.message
        for target in args:
            context.validate_exists(target)
        self.targets = args
        
    def command_parts(self):
        parts = ["commit"]
        message = self.message
        if message is not None:
            parts.extend(["-m", message])
        targets = self.targets
        if targets:
            parts.extend(targets)
        return parts
    
class WithTargets(BaseCommand):
    """Base class for commands that take a set of target files"""
    targets_required = False
    
    def __init__(self, context, args):
        super(WithTargets, self).__init__(context, args)
        for target in args:
            context.validate_exists(target)
            
        if args:
            self.targets = args
        else:
            if self.targets_required:
                raise UVCError("You must list at least one file.")
            else:
                self.targets = None
            
    def command_parts(self):
        parts = [self.__class__.__name__]
        if self.targets:
            parts.extend(self.targets)
        return parts

class diff(WithTargets):
    """The diff command"""
    
    reads_remote = False
    writes_remote = False
    
class remove(WithTargets):
    """Remove a file from the repository."""
    targets_required = True
    
    reads_remote = False
    writes_remote = False
    
class add(WithTargets):
    """The add command, to add files"""
    
    reads_remote = False
    writes_remote = False

class resolved(WithTargets):
    """Marks files as resolved."""
    
    reads_remote = False
    writes_remote = False
    
class status(WithTargets):
    """Retrieve the status of the files in the working copy."""
    
    reads_remote = False
    writes_remote = False

class log(WithTargets):
    """Retrieve the log of changes to a file/project."""

    reads_remote = False
    writes_remote = False

class cat(BaseCommand):
    """Retrieve a specific revision of a file."""
    reads_remote = False
    writes_remote = False

    parser = OptionParser()
    parser.add_option('-r', "--revision", dest="revision", help="revision identifier")

    def __init__(self, context, args):
        super(cat, self).__init__(context, args)
        options, args = self.parser.parse_args(args)
        self.revision = options.revision
        for target in args:
            context.validate_exists(target)
        self.targets = args

    def command_parts(self):
        parts = [ "cat" ]
        revision = self.revision
        if revision is not None:
            parts.extend([ "-r", revision ])
        targets = self.targets
        if targets:
            parts.extend(targets)
        return parts

class revert(WithTargets):
    """Revert a set of files"""
    
    reads_remote = False
    writes_remote = False

class push(BaseCommand):
    """The push command"""
    
    reads_remote = True
    writes_remote = True
    
    def __init__(self, context, args):
        super(push, self).__init__(context, args)
        
        if args:
            dest = args[0]
            if context.remote_scheme_whitelist:
                parsed = urlparse(dest)
                if parsed.scheme not in context.remote_scheme_whitelist:
                    raise SecurityError("Push is limited to only permitted remote repositories")
            self.dest = _apply_auth(dest, context)
        else:
            self.dest = None
    
    def command_parts(self):
        parts = ["push"]
        if self.dest:
            parts.append(self.dest)
            
        return parts

class update(BaseCommand):
    """The update command, intended to act like Mercurial's fetch:
    does a pull from the remote repository and an update of the local
    working copy, merging as needed."""
    
    reads_remote = True
    writes_remote = False
    
    source = None
    
    def __init__(self, context, args):
        super(update, self).__init__(context, args)
        if args:
            source = args[0]
            parsed = urlparse(source)
            
            if context.remote_scheme_whitelist:
                if parsed.scheme not in context.remote_scheme_whitelist:
                    raise SecurityError("Push is limited to only permitted remote repositories")

            source = _apply_auth(source, context)
            self.source = source
    
    def command_parts(self):
        parts = ["update"]
        if self.source:
            parts.append(self.source)
        return parts
    
def get_command_class(context, args):
    """Retrieve a command's class."""
    command_name = args.pop(0).lower()
    command_class = globals().get(command_name)
    if not command_class:
        raise UnknownCommand("There is no %s command." % (command_name))
    return command_class
    
def get_command(context, args):
    """Retrieve a command object based on the command line given"""
    command_class = get_command_class(context, args)
    return command_class(context, args)
    



"""Utility functions used by uvc."""

import os
from subprocess import PIPE, STDOUT
from subprocess import Popen as StdPopen

from uvc.killableprocess import Popen

def run_in_directory(working_dir, command_line, timeout=-1, output=None):
    # use the standard Popen so that we don't have
    # threading issues if there is a threaded
    # server in front
    if timeout == -1 or timeout is None:
        p = StdPopen(command_line, stdout=PIPE, stderr=STDOUT, cwd=working_dir)
    else:
        p = Popen(command_line, stdout=PIPE, stderr=STDOUT, cwd=working_dir)
    while True:
        o = p.stdout.readline()
        if o == '' and p.poll() != None:
            break
        output.write(o)

    output.return_code = p.returncode



"""Exceptions for UVC"""

class UVCError(Exception):
    """Root exception for UVC problems"""
    pass
    
class BadArgument(UVCError):
    """An argument provided to the command was invalid."""
    pass

class UnknownCommand(UVCError):
    """Error when a non-existent command is requested."""
    pass

class FileError(UVCError):
    """Error for files and directories that are out of place."""
    pass

class SecurityError(UVCError):
    """Error for attempting to access something that is
    not known to be secure for remote access."""
    pass
    
class RepositoryAlreadyInitialized(UVCError):
    """Denotes that a directory is already version
    controlled when the user tries to initialize
    it."""
    pass


"""Implements the Mercurial VCS dialect."""
import os
import re

from uvc.commands import UVCError, DialectCommand, BaseCommand
from uvc.exc import RepositoryAlreadyInitialized

class HgError(UVCError):
    """A Mercurial-dialect specific error."""
    pass

class HgCommand(DialectCommand):
    dialect_name = "hg"

class AuthHgCommand(HgCommand):
    def command_parts(self):
        parts = super(AuthHgCommand, self).command_parts()
        auth = self.generic.auth
        if auth:
            if auth['type'] == "ssh":
                parts.insert(1, "--ssh")
                # setting StrictHostKeyChecking to no is less than ideal...
                parts.insert(2, "ssh -i %s -o StrictHostKeyChecking=no" % (auth['key']))
        return parts

class init(HgCommand):
    reads_remote = False
    writes_remote = False
    
    @classmethod
    def from_args(cls, context, args):
        return cls(context, args)
        
    def __init__(self, context, args):
        from uvc import main
        dialect = main.infer_dialect(context.working_dir)
        if dialect:
            raise RepositoryAlreadyInitialized("It looks like there is an %s" 
                            " repository there" % dialect.name)
        
    def command_parts(self):
        return ['init']
    

class clone(AuthHgCommand):
    reads_remote = True
    writes_remote = False

checkout = clone

class commit(HgCommand):
    reads_remote = False
    writes_remote = False
    
    def command_parts(self):
        parts = super(commit, self).command_parts()
        if self.generic.user:
            parts.insert(1, "-u")
            parts.insert(2, self.generic.user)
        return parts
    
class diff(HgCommand):
    reads_remote = False
    writes_remote = False
    
class remove(HgCommand):
    reads_remote = False
    writes_remote = False
    
    def command_parts(self):
        parts = super(remove, self).command_parts()
        # uvc's remove command implies "force"
        parts.insert(1, "-f")
        return parts

class add(HgCommand):
    reads_remote = False
    writes_remote = False
    
class push(AuthHgCommand):
    reads_remote = True
    writes_remote = True
    
class update(AuthHgCommand):
    reads_remote = True
    writes_remote = False
    
    def command_parts(self):
        parts = super(update, self).command_parts()
        parts[0] = "fetch"
        return parts

class resolved(HgCommand):
    reads_remote = False
    writes_remote = False
    
    def command_parts(self):
        parts = super(resolved, self).command_parts()
        parts[0] = "resolve"
        parts.insert(1, "-m")
        if not self.targets:
            parts.append("-a")
        return parts
        
class status(HgCommand):
    reads_remote = False
    writes_remote = False

class log(HgCommand):
    """ hg log describes the history of a file """
    reads_remote = False
    writes_remote = False

    def command_parts(self):
        """ Add a --template option for easier to format output """
        parts = super(log, self).command_parts()
        parts.append("--template")
        parts.append("changeset: {node|short}\n" +
                     "user: {author}\n" +
                     "date: {date|isodatesec}\n" +
                     "summary: {desc|firstline}\n" +
                     "\n")
        return parts

    def get_structured_data(self, output):
        """ Convert the output from 'hg log' into a structured output """
        list = []
        grouped_lines = ""

        def process_group(lines):
            data = { }
            # When we've got a group, parse key:value off each line and
            # place the results in dict in the main list
            for gline in grouped_lines.splitlines():
                if gline.find(":") != -1:
                    key, ignore, value = gline.partition(":")
                    data[key.strip()] = value.strip()
            list.append(data)

        for line in output.splitlines():
            # Group lines together breaking on ^changeset:
            if not re.search("^changeset: ", line) or len(grouped_lines) == 0:
                grouped_lines += "\n" + line
            else:
                process_group(grouped_lines)
                # This line was part of the next group
                grouped_lines = line

        # Catch the stragglers
        process_group(grouped_lines)

        return list

class cat(HgCommand):
    reads_remote = False
    writes_remote = False

class revert(HgCommand):
    reads_remote = False
    writes_remote = False
    
    def command_parts(self):
        parts = super(revert, self).command_parts()
        parts.insert(1, "--no-backup")
        if not self.targets:
            parts.append("-a")
        return parts
    

class HgDialect(object):
    
    name = "hg"
    
    def convert(self, command_object):
        """Converts to a Mercurial-specific command."""
        local_command = self.get_dialect_command_class(command_object.__class__.__name__)
        return local_command(command_object)
    
    def convert_class(self, command_class):
        """Converts a generic command class to its equivalent."""
        local_command_class = self.get_dialect_command_class(command_class.__name__)
        return local_command_class
    
    def get_dialect_command_class(self, command_name):
        """Looks up a Mercurial-specific command and returns the
        class."""
        local_command = globals().get(command_name)
        if not local_command:
            raise HgError("Command cannot be converted to hg: %s" 
                            % (command_name))
        return local_command
    
    
    def cwd_is_this_dialect(self):
        """Returns 1 if the .hg directory is here, 0 otherwise."""
        if os.path.isdir(".hg"):
            return 1
        return 0
    
